import rhinoscriptsyntax as rhino
import math

def mesh_in_parameter_space_to_quads(meshes = None, scaling_factor = None):
    if scaling_factor == None:
        scaling_factor = 1.0/float(rhino.GetReal(message = "number of Rhino units to one parameter unit", number = 10))
    faces = []
    for mesh in meshes:
        meshfaces = rhino.MeshFaces(mesh) #this is a list of length 4n, where n is the number of tris and quads
        for i in range(int(len(meshfaces)/4)): 
            face = meshfaces[4*i:4*i+4]
            face = [[scaling_factor*coord for coord in vertex] for vertex in face]
            faces.append(face)
    return faces

def quadrilateral_linear_interpolate(u, v, corners):
    aa, ab, bb, ba = corners #corners ordered cyclically!
    aa0, aa1, aa2 = aa
    ab0, ab1, ab2 = ab
    bb0, bb1, bb2 = bb
    ba0, ba1, ba2 = ba
    out0 = (aa0*(1-u) + ba0*u) * (1-v) + (ab0*(1-u) + bb0*u) * v
    out1 = (aa1*(1-u) + ba1*u) * (1-v) + (ab1*(1-u) + bb1*u) * v
    out2 = (aa2*(1-u) + ba2*u) * (1-v) + (ab2*(1-u) + bb2*u) * v
    return [out0, out1, out2]
    #if triangle then bb==ba, then u==1 gives constant point

def pointsarray_from_mesh_face(face, num_points):
    patch = []
    for i in range(num_points):
            u = float(i)/float(num_points-1)
            row = []
            for j in range(num_points):
                    v = float(j)/float(num_points-1)
                    point = quadrilateral_linear_interpolate(u, v, face)
                    row.append(point)
            patch.append(row) # note that a patch will repeat points when
                                                # the face is a triangle
    return patch

def draw_parametric_surface_patches_from_pointsarrays(pointsarrays, draw_surfaces = True, draw_points = False, scale = 10.0):
    rhino.EnableRedraw(False) #faster if we don't redraw every time we make a new patch
    surfaces_out = []
    points_out = []
    for pointsarray in pointsarrays:
        pointsarray = [[[scale*coord for coord in point] for point in row] for row in pointsarray]
        if draw_surfaces:
            num_points_u = len(pointsarray)
            num_points_v = len(pointsarray[0])
            points = []
            for row in pointsarray:
                points.extend(row)
            surfaces_out.append(rhino.AddSrfControlPtGrid((num_points_u, num_points_v), points))
        if draw_points:
            for row in pointsarray:
                for point in row:
                    points_out.append(rhino.AddPoint(point))
    rhino.EnableRedraw(True)
    return surfaces_out    

def transform_mesh(function, input_meshes = None, num_points = 5, join_patches = True):
    if input_meshes == None:
            input_meshes = rhino.GetObjects(message = "Select mesh", filter=32, preselect=True) #meshes only
    faces = mesh_in_parameter_space_to_quads(meshes = input_meshes)
    pointsarrays = [pointsarray_from_mesh_face(face, num_points = num_points) for face in faces]
    pointsarrays = [[[function(point[0], point[1], point[2]) for point in row] for row in pointsarray] for pointsarray in pointsarrays]
    patches = draw_parametric_surface_patches_from_pointsarrays(pointsarrays, draw_surfaces = True, draw_points = False, scale = 10.0)
    # print patches
    rhino.JoinSurfaces(patches, delete_input = True)

def spherical_coordinates(rho,theta,phi):
    return [rho*math.cos(theta)*math.sin(phi), rho*math.sin(theta)*math.sin(phi), rho*math.cos(phi)]

transform_mesh(spherical_coordinates)


